/** The N2A03 class emulates the NMOS 6502 microprocessor produced by Ricoh and used in the NES. <BR>
    The 2a03 is a 6502 CPU that does not support the decimal mode of the ADC and SBC instructions.<BR>
    This code is GPL.
    Copyright (c) 2000 Brice Fines.<DL>
    <UR>
    <LI>This program is free software; you can redistribute it and/or modify<BR>
    it under the terms of the GNU General Public License as published by<BR>
    the Free Software Foundation; either version 2 of the License, or<BR>
    (at your option) any later version.<BR>
    <LI>This program is distributed in the hope that it will be useful,<BR>
    but WITHOUT ANY WARRANTY; without even the implied warranty of<BR>
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the<BR>
    GNU General Public License for more details.<BR>
    <LI>You should have received a copy of the GNU General Public License<BR>
    along with this program; if not, write to the Free Software<BR>
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA<BR>
    </UR></DL>
    @version 0.2005 06/06/00
    @author <A HREF="mailto:bfines@uni2.es">Jarod CANAL</A>
    @see "The Gnu Public License." */
public abstract class N2A03 extends NMOS6502{

    /** Default clock. <BR>
	(21477272.724 / 12). */
    private final double N2A03_DEFAULTCLOCK = 1789772.727;

    /** Constructor */
    public N2A03(){
    }

    /** Add with carry. */
    final protected void ADC(){
	final int c = ( processorStatus.get() & F_C );
	final int sum = accumulator.get() + tmp + c;
	processorStatus.set( processorStatus.get() & ~( F_V | F_C ) );
	if ( ( ~( accumulator.get() ^ tmp ) & ( accumulator.get() ^ sum ) & F_N ) != 0 )
	    {
		processorStatus.set( processorStatus.get() | F_V );
	    }
	if ( ( sum & 0xFF00 ) != 0 )
	    {
		processorStatus.set( processorStatus.get() | F_C );
	    }
	accumulator.set( sum );
	setNZ( accumulator.get() );
    }

    /** Subtract with carry. */
    final protected void SBC(){
	final int c = ( processorStatus.get() & F_C ) ^ F_C;
	final int sum = accumulator.get() - tmp - c;
	processorStatus.set( processorStatus.get() & ~( F_V | F_C ) );
	if ( ( ( accumulator.get() ^ tmp ) & ( accumulator.get() ^ sum ) & F_N ) != 0 )
	    {
		processorStatus.set( processorStatus.get() | F_V );
	    }
	if ( ( sum & 0xFF00 ) == 0 )
	    {
		processorStatus.set( processorStatus.get() | F_C );
	    }
	accumulator.set( sum );
	setNZ( accumulator.get() );
    }
    
    /** N2A03 IRQ. <BR>
	The N2A03 is integrally tied to its PSG (they're on the same die). Bit 7 of address $4011 (the PSG's DPCM control register), when set, causes an IRQ to be generated.  This function allows the IRQ to be called from the PSG core when such an occasion arises. */
    final protected void n2a03_irq()
    {
	take_irq();
    }
    
}
