import java.awt.event.*;

/** The NintendoEntertainmentSystem class emulates the Nintendo Entertainment System produced by Nintendo. <BR>
    This class needs a lot of modifications and fixes (lots of bugs). <BR>
    This code is GPL.
    Copyright (c) 2000 Brice Fines.<DL>
    <UR>
    <LI>This program is free software; you can redistribute it and/or modify<BR>
    it under the terms of the GNU General Public License as published by<BR>
    the Free Software Foundation; either version 2 of the License, or<BR>
    (at your option) any later version.<BR>
    <LI>This program is distributed in the hope that it will be useful,<BR>
    but WITHOUT ANY WARRANTY; without even the implied warranty of<BR>
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the<BR>
    GNU General Public License for more details.<BR>
    <LI>You should have received a copy of the GNU General Public License<BR>
    along with this program; if not, write to the Free Software<BR>
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA<BR>
    </UR></DL>
    @version 0.2005 06/06/00
    @author <A HREF="mailto:bfines@uni2.es">Jarod CANAL</A>
    @see "The Gnu Public License." */
final public class NintendoEntertainmentSystem extends N2A03{
    
    /** Cart ROM image. */
    protected char RomImage[];
    
    /** Cart VROM image. */    
    protected char VRomImage[];

    /** NES memory. <BR>
	0x10000 bytes of RAM. */
    protected char Memory[] = new char[ 0x10000 ];
    
    /** NES picture processing unit. */
    protected Video PPU = new Video();
    
    /** Joypad 1 variable. */
    protected int joy_1;
    
    /** Joypad 2 variable. */
    protected int joy_2;
    
    /** Joypad 1 bit variable. */
    protected int joy_readbit1;
    
    /** Joypad 2 bit variable. */
    protected int joy_readbit2;
    
    /** Joypad variable. <BR>
	If $01 is written to $4016 then this value is set to true. */
    private boolean reset_joy = false;

    /** Constructor. */
    public NintendoEntertainmentSystem(){}
    
    /** Constructor.
	@param RomPages number of ROM pages.
	@param VRomPages number of VROM pages.
	@param HasHorizontalMirroring indicates if cart has horizontal or vertical mirroring. */
    public NintendoEntertainmentSystem( int RomPages, int VRomPages, boolean HasHorizontalMirroring ){
	RomImage = new char[ RomPages * 0x4000 ];
	VRomImage = new char[ VRomPages * 0x2000 ];
	if( HasHorizontalMirroring )
	    { // fixme move to ppu class
		PPU.NameTable1 = 0x2000;
		PPU.NameTable2 = 0x2400;
	    }
	else
	    {
		PPU.NameTable1 = 0x2400;
		PPU.NameTable2 = 0x2000;
	    }
	PPU.canvas.addKeyListener(new KeyAdapter()
	    {
		final public void keyPressed(KeyEvent e)
		{
		    switch( e.getKeyCode() )
			{
			case java.awt.event.KeyEvent.VK_D : // A
			    joy_1 |= 0x01;
			    break;
			case java.awt.event.KeyEvent.VK_S : // B
			    joy_1 |= 0x02;
			    break;
			case java.awt.event.KeyEvent.VK_F : // Select
			    joy_1 |= 0x04;
			    break;
			case java.awt.event.KeyEvent.VK_G : // Start
			    joy_1 |= 0x08;
			    break;
			case java.awt.event.KeyEvent.VK_UP : // Up
			    joy_1 |= 0x10;
			    break;
			case java.awt.event.KeyEvent.VK_DOWN : // Down
			    joy_1 |= 0x20;
			    break;
			case java.awt.event.KeyEvent.VK_LEFT : // Left
			    joy_1 |= 0x40;
			    break;
			case java.awt.event.KeyEvent.VK_RIGHT : // Right
			    joy_1 |= 0x80;
			    break;
			default :
			    break;
			}
		}
		final public void keyReleased(KeyEvent e)
		{
		    switch( e.getKeyCode() )
			{
			case java.awt.event.KeyEvent.VK_D : // A
			    joy_1 &= ~0x01;
			    break;
			case java.awt.event.KeyEvent.VK_S : // B
			    joy_1 &= ~0x02;
			    break;
			case java.awt.event.KeyEvent.VK_F : // Select
			    joy_1 &= ~0x04;
			    break;
			case java.awt.event.KeyEvent.VK_G : // Start
			    joy_1 &= ~0x08;
			    break;
			case java.awt.event.KeyEvent.VK_UP : // Up
			    joy_1 &= ~0x10;
			    break;
			case java.awt.event.KeyEvent.VK_DOWN : // Down
			    joy_1 &= ~0x20;
			    break;
			case java.awt.event.KeyEvent.VK_LEFT : // Left
			    joy_1 &= ~0x40;
			    break;
			case java.awt.event.KeyEvent.VK_RIGHT : // Right
			    joy_1 &= ~0x80;
			    break;
			default :
			    break;
			}
		}
	    });
    }

    public void WRMEM( int Address, int Value ){
	// pending add a sanity check
	switch( Address & 0xE000 )
	    {
		// RAM access
	    case 0x0000 :
		Memory[ Address & 0x07FF ] = ( char ) Value;
		break;
		// IO access
	    case 0x2000 :
		switch ( Address & 7 )
		    {
		    case 0:
			Memory[ 0x2000 ] = ( char ) Value;
			PPU.TempVRAMAddressRegister &= 0x73FF;
			PPU.TempVRAMAddressRegister |= ( Value & 0x03 ) << 10;
			if( ( Value & 0x04 ) == 0 )
			    {
				PPU.VRAMAddressIncrement = 1;
			    }
			else
			    {
				PPU.VRAMAddressIncrement = 32;
			    }
			if( ( Value & 0x08 ) == 0 )
			    {
				PPU.SpritePatternTableAddress = 0;
			    }
			else
			    {
				PPU.SpritePatternTableAddress = 0x100; // pending 256 to optimize cache therefore move it in Video
			    }
			if( ( Value & 0x10 ) == 0 )
			    {
				PPU.BackgroundPatternTableAddress = 0;
			    }
			else
			    {
				PPU.BackgroundPatternTableAddress = 0X100; // pending 256 to optimize cache therefore move it in Video
			    }
			if( ( Value & 0x20 ) == 0 )
			    {
				PPU.SpriteSize = 8;
			    }
			else
			    {
				PPU.SpriteSize = 16;
			    }
			if( ( Value & 0x40 ) == 0 )
			    {
				PPU.IsPPUSlave = false;
			    }
			else
			    {
				PPU.IsPPUSlave = true;
			    }
			if( ( Value & 0x80 ) == 0 )
			    {
				PPU.IsExecuteNMIOnVBlankEnabled = false;
			    }
			else
			    {
				PPU.IsExecuteNMIOnVBlankEnabled = true;
			    }
			break;
		    case 1:
			Memory[ 0x2001 ] = ( char ) Value;
			if( ( Value & 0x01 ) == 0 )
			    {
				PPU.IsDisplayTypeColorDisplay = true;
			    }
			else
			    {
				PPU.IsDisplayTypeColorDisplay = false;
			    }
			if( ( Value & 0x02 ) == 0 )
			    {
				PPU.BackgroundClipping = 8;
			    }
			else
			    {
				PPU.BackgroundClipping = 0;
			    }
			if( ( Value & 0x04 ) == 0 )
			    {
				PPU.SpriteClipping = 8;
			    }
			else
			    {
				PPU.SpriteClipping = 0;
			    }
			if( ( Value & 0x08 ) == 0 )
			    {
				PPU.BackgroundVisibility = false;
			    }
			else
			    {
				PPU.BackgroundVisibility = true;
			    }
			if( ( Value & 0x10 ) == 0 )
			    {
				PPU.SpriteVisibility = false;
			    }
			else
			    {
				PPU.SpriteVisibility = true;
			    }
			if( PPU.IsDisplayTypeColorDisplay )
			    {
				switch( Value & 0x00E0 )
				    {
				    case 0x00 :
					PPU.ColorIntensity = 0;
					break;
				    case 0x20 :
					PPU.ColorIntensity = 1;
					break;
				    case 0x40 :
					PPU.ColorIntensity = 2;
					break;
				    case 0x80 :
					PPU.ColorIntensity = 3;
					break;
				    default : System.out.println( "Invalid value for color intensity." );
				    }
			    }
			else{
			    switch( Value & 0x00E0 )
				{
				case 0x00 :
				    PPU.FullBackgroundColor = 0; // pending replace with RGB values
				    break;
				case 0x20 :
				    PPU.FullBackgroundColor = 1;
				    break;
				case 0x40 :
				    PPU.FullBackgroundColor = 2;
				    break;
				case 0x80 :
				    PPU.FullBackgroundColor = 3;
				    break;
				default : System.out.println( "Invalid value for full background color." );
				}
			}
			break;
		    case 2:
			System.out.println( "ppu_io_write : $2002 is not writable." );
			break;
		    case 3:
			Memory[ 0x2003 ] = ( char ) Value;
			break;
		    case 4:
			Memory[ 0x2004 ] = ( char ) Value;
			PPU.SpriteRam[ Memory[ 0x2003 ] ] = ( char ) Value;
			Memory[ 0x2003 ] ++;
			Memory[ 0x2003 ] &= 0x00FF; // Memory wrap
			break;
		    case 5:
			Memory[ 0x2005 ] = ( char ) Value;
			if( PPU.AddressMode )
			    { // second write
				PPU.AddressMode = false;
				PPU.TempVRAMAddressRegister &= 0x0C1F;
				PPU.TempVRAMAddressRegister |= ( ( Value & 7 ) << 12 );
				PPU.TempVRAMAddressRegister |= ( ( Value << 2 ) & 0x03E0 );
			    }
			else
			    { // first write
				PPU.AddressMode = true;
				PPU.TempVRAMAddressRegister &= 0x7FE0;
				PPU.TempVRAMAddressRegister |= ( Value >> 3 );
				PPU.XOffset = ( Value & 7 );
			    }
			break;
		    case 6:
			Memory[ 0x2006 ] = ( char ) Value;
			if( PPU.AddressMode )
			    { // second write
				PPU.AddressMode = false;
				PPU.VRAMAddress = ( char ) ( ( PPU.HalfVRAMAddress << 8 ) | Value );
				PPU.TempVRAMAddressRegister &= 0xFF00;
				PPU.TempVRAMAddressRegister |= Value;
				PPU.VRAMAddressRegister = PPU.TempVRAMAddressRegister;
			    }
			else
			    { // first write
				PPU.AddressMode = true;
				PPU.HalfVRAMAddress = ( char ) Value;
				PPU.TempVRAMAddressRegister &= 0x00FF;
				PPU.TempVRAMAddressRegister |= ( ( Value & 0x3F ) << 8 );
			    }
			break;
		    case 7:
			Memory[ 0x2007 ] = ( char ) Value;
			if( PPU.AddressMode )
			    {
				System.out.println( "PPU write during address sequence." );
			    }
			PPU.VRAMAddress &= 0x03FFF;
			switch( PPU.VRAMAddress & 0x3000 )
			    {
			    case 0x0000 :
			    case 0x1000 :
				break; // Cannot write to VROM, pending add VROM 0K support
			    case 0x2000 :
				switch( PPU.VRAMAddress & 0x0F00 )
				    {
				    case 0x000 :
				    case 0x100 :
				    case 0x200 :
				    case 0x300 :
					PPU.Memory[ PPU.VRAMAddress ] = ( char ) Value;
					break;
				    case 0x400 :
				    case 0x500 :
				    case 0x600 :
				    case 0x700 :
					PPU.Memory[ PPU.NameTable1 + ( PPU.VRAMAddress & 0x3FF ) ] = ( char ) Value;
					break;
				    case 0x800 :
				    case 0x900 :
				    case 0xA00 :
				    case 0xB00 :
					PPU.Memory[ PPU.NameTable2 + ( PPU.VRAMAddress & 0x3FF ) ] = ( char ) Value;
					break;
				    case 0xC00 :
				    case 0xD00 :
				    case 0xE00 :
				    case 0xF00 :
					PPU.Memory[ 0x2400 + ( PPU.VRAMAddress & 0x3FF ) ] = ( char ) Value;
				    }
				break;
			    case 0x3000 :
				if( PPU.VRAMAddress > 0x3EFF )
				    {// pending move palette caching
					if( ( PPU.VRAMAddress & 0x3FEF ) == 0x3F00 )
					    {
						PPU.Memory[ 0x3F00 ] = PPU.Memory[ 0x3F04 ] = PPU.Memory[ 0x3F08 ] = PPU.Memory[ 0x3F0C ] = PPU.Memory[ 0x3F10 ] = PPU.Memory[ 0x3F14 ] = PPU.Memory[ 0x3F18 ] = PPU.Memory[ 0x3F1C ] = ( char ) Value;
						PPU.PaletteCache[ 0x00 ] = PPU.PaletteCache[ 0x04 ] =  PPU.PaletteCache[ 0x08 ] =  PPU.PaletteCache[ 0x0C ] =  PPU.PaletteCache[ 0x10 ] =  PPU.PaletteCache[ 0x14 ] =  PPU.PaletteCache[ 0x18 ] = PPU.PaletteCache[ 0x1C ] = PPU.rgb[ Value & 0x003F ];
					    }
					else
					    {
						PPU.Memory[ 0x3F00 + ( PPU.VRAMAddress & 0x001F ) ] = ( char ) Value;
						PPU.PaletteCache[ PPU.VRAMAddress & 0x001F ] = PPU.rgb[ Value & 0x003F ]; // pending move it, optimize PPU.VRAMAddress & 0x001F
					    }
				    }
			    }
			PPU.VRAMAddress += PPU.VRAMAddressIncrement;
		    }
		break;
	    case 0x4000 :
		if( Address < 0x4018 )
		    {
			Memory[ Address ] = ( char ) Value;
			switch( Address )
			    {
			    case 0x4000 :
				break;
			    case 0x4001 :
				break;
			    case 0x4002 :
				break;
			    case 0x4003 :
				break;
			    case 0x4004 :
				break;
			    case 0x4005 :
				break;
			    case 0x4006 :
				break;
			    case 0x4007 :
				break;
			    case 0x4008 :
				break;
			    case 0x4009 :
				break;
			    case 0x400A :
				break;
			    case 0x400B :
				break;
			    case 0x400C :
				break;
			    case 0x400E :
				break;
			    case 0x400F :
				break;
			    case 0x4010 :
				break;
			    case 0x4011 :
				break;
			    case 0x4012 :
				break;
			    case 0x4013 :
				break;
			    case 0x4014 :
				switch( Value & 0xF0 )
				    {
				    case 0x20 :
				    case 0x30 :
				    case 0x40 :
				    case 0x50 :
					System.out.println( "Ignoring sprite DMA from registers and expansion zones." );
					break;
				    default : System.arraycopy( Memory, Value << 8, PPU.SpriteRam, 0, 0x100 );
				    }
				break;
			    case 0x4015 :
				break;
			    case 0x4016 : // Joypad #1 write 
				if( reset_joy && ( Value == 0 ) )
				    { // Reset.
					joy_readbit1 = 0;
					joy_readbit2 = 0;
					reset_joy = false;
				    }	
				else
				    {
					reset_joy = false;
				    }
				if( Value == 1 )
				    {
					reset_joy = true;
				    }
				break;
			    case 0x4017 : // Joypad #2 write, shouldn't happen, doesn't do anything
				break;
			    default : break;
			    }
		    }
		break;
		// SRAM access //pending modify
	    case 0x6000 :
		Memory[ Address ] = ( char ) Value;
		break;
		// fixme add case 0x8000 to 0XE000
	    default : break; // Address is not writable
	    }
    }
    
    public int RDMEM( int Address ){
	// pending add a sanity check
	switch( Address & 0xE000 )
	    {
		// RAM access
	    case 0x0000 :
		return( Memory[ Address & 0x07FF ] );
		// IO access
	    case 0x2000 :
		if( ( Address & 7 ) == 7 )
		    { // pending switch more efficient??
			PPU.VRAMAddress &= 0x03FFF;
			char Result; 
			if( Address > 0x3EFF )
			    {
				Result = PPU.Memory[ 0x3F00 + ( PPU.VRAMAddress & 0x001F ) ];
			    }
			else
			    {
				Result = PPU.VRAMDataBuffer;
				switch( PPU.VRAMAddress & 0x0F00 )
				    {
				    case 0x000 :
				    case 0x100 :
				    case 0x200 :
				    case 0x300 : PPU.VRAMDataBuffer = PPU.Memory[ PPU.VRAMAddress ];
					break;
				    case 0x400 :
				    case 0x500 :
				    case 0x600 :
				    case 0x700 : PPU.VRAMDataBuffer = PPU.Memory[ PPU.NameTable1 + ( PPU.VRAMAddress & 0x3FF ) ];
					break;
				    case 0x800 :
				    case 0x900 :
				    case 0xA00 :
				    case 0xB00 : PPU.VRAMDataBuffer = PPU.Memory[ PPU.NameTable2 + ( PPU.VRAMAddress & 0x3FF ) ];
					break;
				    case 0xC00 :
				    case 0xD00 :
				    case 0xE00 :
				    case 0xF00 : PPU.VRAMDataBuffer = PPU.Memory[ 0x2400 + ( PPU.VRAMAddress & 0x3FF ) ];
				    }
			    }
			PPU.VRAMAddress += PPU.VRAMAddressIncrement;
			return( Result );
		    }
		else
		    {
			char Result2 = Memory[ 0x2002 ];
			Memory[ 0x2002 ] &= 0x007F;
			PPU.AddressMode = false;
			return( Result2 );
		    }
	    case 0x4000 :
		switch( Address )
		    {
		    case 0x4016 : 
			joy_readbit1 &= 7; // to be modified to add signature support
			return( ( joy_1 >> ( joy_readbit1 ++ ) ) & 1 );
		    case 0x4017 :
			joy_readbit2 &= 7; // to be modified to add signature support
			return( ( joy_2 >> ( joy_readbit2 ++ ) ) & 1 );
		    default : return( ( char ) ( Address >> 8 ) ); // Address is not readable
		    }
		//pending SRAM support
	    default : return( Memory[ Address ] ); //Address is readable
	    }
    }
    
    protected void nes_timeslice(){
	for (;;)
	    {
		execute(113);
		if ( PPUPeriodic() )
		    {
			set_nmi_line( ( byte ) 1 );
		    }
		else
		    {
			set_nmi_line( 0 );
		    }
	    }
    }
    
    protected boolean PPUPeriodic(){
	if( PPU.Scanline < 240 )
	    {
		//if( PPU.Scanline == PPU.SpriteRam[0] + 6 ) {PPU.WasSpriteZeroHit=true;/*Memory[0x2002] |= 0x0040;*/} // not here in darcnes hit flag set
		/*if( PPU.BackgroundVisibility && PPU.SpriteVisibility )
		  {*/
		PPU.VRAMAddressRegister &= 0x7BE0;
		PPU.VRAMAddressRegister |= ( PPU.TempVRAMAddressRegister & 0x041F );
		//}
		PPU.drawLine();
		if( PPU.WasSpriteZeroHit ) Memory[0x2002] |= 0x40;
		if( PPU.IsThereMoreThanEightSpritesOnScanline ) Memory[0x2002] |= 0x20;
	    }
	else
	    {
		if( PPU.Scanline == 243 )
		    { // in darcness vblank is on 240
			Memory[ 0x2002 ] = 0x0080;
			PPU.WasSpriteZeroHit = false;
			PPU.IsThereMoreThanEightSpritesOnScanline = false;
			PPU.drawDisplay();
		    }
	    }
	PPU.Scanline ++;
	if( PPU.Scanline == 262 )
	    {
		PPU.Scanline = 0;
		Memory[ 0x2002 ] = 0;
		PPU.WasSpriteZeroHit = false;
		PPU.IsThereMoreThanEightSpritesOnScanline = false;
		/*if( PPU.BackgroundVisibility && PPU.SpriteVisibility )
		  {*/
		PPU.VRAMAddressRegister = PPU.TempVRAMAddressRegister;
		//}
	    }
	return ( ( PPU.IsExecuteNMIOnVBlankEnabled ) && ( PPU.Scanline == 243 ) );
    }
    
}
