/** This class provides an union structure to port Juergen's C code.<BR>
    The union can contain a double word, a word or a byte.<BR>
    0xFFFFFFFF.<BR>
    This code is GPL. 
    Copyright (c) 2000 Brice Fines.<DL>
    <UR>
    <LI>This program is free software; you can redistribute it and/or modify<BR>
    it under the terms of the GNU General Public License as published by<BR>
    the Free Software Foundation; either version 2 of the License, or<BR>
    (at your option) any later version.<BR>
    <LI>This program is distributed in the hope that it will be useful,<BR>
    but WITHOUT ANY WARRANTY; without even the implied warranty of<BR>
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the<BR>
    GNU General Public License for more details.<BR>
    <LI>You should have received a copy of the GNU General Public License<BR>
    along with this program; if not, write to the Free Software<BR>
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA<BR>
    </UR></DL>
    @version 0.2005 06/06/00<BR>
    @author <A HREF="mailto:bfines@uni2.es">Jarod CANAL</A>
    @see "The Gnu Public License." */
final public class Pair{

    /** A table of 4 unsigned 8 bits integers that represents the union.<BR>
	0xFF FF FF FF */
    private int unsignedByte;
    
    /** Constructor. */
    public Pair(){}
    
    /** Set the union with a double word value.<BR>
	0xFF FF FF FF
	@param Value the value that will be used to set the union.*/
    final protected void setDWord( int Value ){
	unsignedByte = Value;
    }
    
    /** Set the union with a word value.<BR>
	0x 00 00 FF FF
	@param Value the value that will be used to set the union.*/
    final protected void setWord( int Value ){
	unsignedByte &= 0xFFFF0000;
	unsignedByte |= ( Value & 0xFFFF ); // Is it unsafe to quit & 0xFFFF ?
    }
    
    /** Set the low byte of the union with a value.<BR>
	0x 00 00 00 FF
	@param Value the value that will be used to set the union. */
    final protected void setLowByte( int Value ){
	unsignedByte &= 0xFFFFFF00;
	unsignedByte |= ( Value & 0xFF ); // Is it unsafe to quit & 0xFF ?
    }
    
    /** Set the high byte of the union with a value.<BR>
	0x 00 00 FF 00
	@param Value the value that will be used to set the union. */
    final protected void setHighByte( int Value ){
	unsignedByte &= 0xFFFF00FF;
	unsignedByte |= ( ( Value & 0xFF ) << 8 ); // Is it unsafe to quit & 0xFF ?
    }
    
    /** Get the double word value of the union.<BR>
	0x FF FF FF FF
	@return an unsigned 32 bits integer.*/
    final protected int getDWord(){
	return( unsignedByte );
    }
    
    /** Get the word value of the union.<BR>
	0x 00 00 FF FF
	@return an unsigned 16 bits integer. */
    final protected int getWord(){
	return( unsignedByte & 0xFFFF );
    }
    
    /** Get the low byte value of the union.<BR>
	0x 00 00 00 FF
	@return an unsigned 8 bits integer. */
    final protected int getLowByte(){
	return( unsignedByte & 0xFF );
    }
    
    /** Get the high byte value of the union.<BR>
	0x 00 00 FF 00
	@return an unsigned 8 bits integer. */
    final protected int getHighByte(){
	return( ( unsignedByte >> 8 ) & 0xFF );
    }
    
}
